#pragma once

#include <memory>

#include <dune/istl/operators.hh>
#include <dune/istl/scalarproducts.hh>
#include <dune/istl/solvercategory.hh>

#if HAVE_MPI
#include <dune/istl/novlpschwarz.hh>
#include <dune/istl/schwarz.hh>
#endif

#include <amdis/Output.hpp>

namespace AMDiS
{
  /// Creator to create ScalarProduct objects
  template <class X>
  class ISTLScalarProductCreator
  {
    using Interface = Dune::ScalarProduct<X>;

  public:
    template <class C>
    static std::unique_ptr<Interface> create(Dune::SolverCategory::Category cat, C const& comm)
    {
      switch (cat)
      {
        case Dune::SolverCategory::Category::sequential :
          return std::make_unique< Dune::SeqScalarProduct<X> >();
#if HAVE_MPI
        case Dune::SolverCategory::Category::overlapping :
          return std::make_unique< Dune::OverlappingSchwarzScalarProduct<X,C> >(comm);
        case Dune::SolverCategory::Category::nonoverlapping :
          return std::make_unique< Dune::NonoverlappingSchwarzScalarProduct<X,C> >(comm);
#endif
        default:
          error_exit("Invalid solver category for scalar product\n");
          return nullptr; // avoid warnings
      }
    }
  };

  /// Creator to create Linear Operator objects
  template <class A, class X, class B>
  class ISTLLinearOperatorCreator
  {
    using Interface = Dune::AssembledLinearOperator<A,X,B>;

  public:
    template <class C>
    static std::unique_ptr<Interface> create(Dune::SolverCategory::Category cat, A const& mat, C const& comm)
    {
      switch (cat)
      {
        case Dune::SolverCategory::Category::sequential :
          return std::make_unique< Dune::MatrixAdapter<A,X,B> >(mat);
#if HAVE_MPI
        case Dune::SolverCategory::Category::overlapping :
          return std::make_unique< Dune::OverlappingSchwarzOperator<A,X,B,C> >(mat, comm);
        case Dune::SolverCategory::Category::nonoverlapping :
          return std::make_unique< Dune::NonoverlappingSchwarzOperator<A,X,B,C> >(mat, comm);
#endif
        default:
          error_exit("Invalid solver category for linear operator\n");
          return nullptr; // avoid warnings
      }
    }
  };

  /// Creator to create parallel Preconditioners
  template <class X, class B>
  class ISTLParallelPreconditionerCreator
  {
    using P = Dune::Preconditioner<X,B>;

  public:
    template <class C>
    static std::unique_ptr<P> create(Dune::SolverCategory::Category cat, P& prec, C const& comm)
    {
      switch (cat)
      {
#if HAVE_MPI
        case Dune::SolverCategory::Category::overlapping :
          return std::make_unique< Dune::BlockPreconditioner<X,B,C> >(prec, comm);
        case Dune::SolverCategory::Category::nonoverlapping :
          return std::make_unique< Dune::NonoverlappingBlockPreconditioner<C,P> >(prec, comm);
#endif
        default:
          error_exit("Invalid solver category for parallel preconditioner\n");
          return nullptr; // avoid warnings
      }
    }
  };

} // end namespace AMDiS
