#pragma once

#include <dune/common/version.hh>
#include <dune/istl/solvercategory.hh>

#include <amdis/linearalgebra/istl/Fwd.hpp>
#include <amdis/linearalgebra/istl/ISTLPreconCreatorInterface.hpp>
#include <amdis/linearalgebra/RunnerInterface.hpp>
#include <amdis/linearalgebra/SolverInfo.hpp>
#include <amdis/Environment.hpp>
#include <amdis/Output.hpp>

namespace AMDiS
{
  template <class ISTLSolver, class Traits>
  class ISTLRunner
      : public RunnerInterface<Traits>
  {
    using Mat = typename Traits::Mat;
    using Sol = typename Traits::Sol;
    using Rhs = typename Traits::Rhs;
    using ISTLPreconCreator  = ISTLPreconCreatorInterface<Mat, Sol, Rhs>;

  public:
    ISTLRunner(std::string const& prefix)
      : solverCreator_(prefix)
    {
      initPrecon(prefix);
    }


    /// Implements \ref RunnerInterface::init()
    void init(Mat const& A, typename Traits::Comm& comm) override
    {
      auto cat = Dune::SolverCategory::category(comm);

      sp_ = Traits::ScalProdCreator::create(cat, comm);
      if (cat == Dune::SolverCategory::Category::sequential)
      {
        precon_ = preconCreator_->create(A);
      }
      else
      {
        basePrecon_ = preconCreator_->create(A);
        precon_ = Traits::ParPrecCreator::create(cat, *basePrecon_, comm);
      }
      linOperator_ = Traits::LinOpCreator::create(cat, A, comm);
      solver_ = solverCreator_.create(*linOperator_, *sp_, *precon_);
    }

    /// Implements \ref RunnerInterface::exit()
    void exit() override
    {
      solver_.reset();
      linOperator_.reset();
      precon_.reset();
      basePrecon_.reset();
      sp_.reset();
    }

    /// Implements \ref RunnerInterface::solve()
    int solve(Mat const& A, Sol& x, Rhs const& b, SolverInfo& solverInfo) override
    {
      DUNE_UNUSED_PARAMETER(A);

      // storing some statistics
      Dune::InverseOperatorResult statistics;

      // solve the linear system
      Rhs _b = b;
      solver_->apply(x, _b, statistics);

      solverInfo.setRelResidual(statistics.reduction);

      return statistics.converged ? 0 : 1;
    }

  protected:
    void initPrecon(std::string const& prefix)
    {
      // get creator string for preconditioner
      std::string preconName = "no";
      std::string initFileStr = "";
      for (std::string postfix : {"left precon", "right precon", "precon", "precon->name"}) {
        Parameters::get(prefix + "->" + postfix, preconName);
        if (!preconName.empty() && preconName != "no") {
          initFileStr = postfix == "precon->name" ? prefix + "->precon" : prefix + "->" + postfix;
          break;
        }
      }

      if (preconName.empty() || preconName == "no")
        preconName = "default";

      // TODO(FM): Find a better test if preconditioner is supported in parallel
      if (Environment::mpiSize() > 1) {
        std::set<std::string> tested_precons = {"jacobi", "diag", "default", "richardson"};
        test_warning(tested_precons.count(preconName) == 1, "Preconditioner {} not supported in parallel.", preconName);
      }

      auto* creator = named(CreatorMap<ISTLPreconCreator>::getCreator(preconName, initFileStr));
      preconCreator_ = creator->createWithString(initFileStr);
      assert(preconCreator_);
    }

  private:
    ISTLSolverCreator<ISTLSolver> solverCreator_;
    std::shared_ptr<ISTLPreconCreator> preconCreator_;

    std::shared_ptr<typename Traits::ScalProd> sp_;
    std::shared_ptr<typename Traits::Prec>     basePrecon_;
    std::shared_ptr<typename Traits::Prec>     precon_;
    std::shared_ptr<typename Traits::LinOp>    linOperator_;
    std::shared_ptr<typename Traits::Solver>   solver_;
  };

} // end namespace AMDiS
